@@ -36,6 +36,15 @@
 #include <cstdio>
 #include <cstdlib>
 #include <cstring> // for strchr()
+#ifdef HAVE_DIRENT_H
+# include <sys/types.h>
+# include <dirent.h>
+# include <errno.h>
+#endif
+#ifdef HAVE_WINDOWS_H
+# include <windows.h>
+# include <tchar.h>
+#endif
 
 // The actual algorithms for number checking are there
 #include "algorithms.h"
@@ -44,7 +53,9 @@
 // are in accnum.cc.
 
 #if defined(__MINGW32__) || defined(_WIN32)
-# define OS_WIN32
+# ifndef OS_WIN32
+#  define OS_WIN32
+# endif
 #endif
 
 #ifdef OS_WIN32
@@ -55,23 +66,15 @@
 # define HAVE_ASCTIME
 #endif
 
-using namespace std;
+#ifdef ERROR
+# undef ERROR
+#endif
 
-// A hard-coded list of available files in the bankdata directory. In
-// the long run, this should be replaced by a dynamic lookup in
-// populate_dated_files(), but for now this is the easiest solution.
-// const char* file_dates[] = {
-//   "bankdata_20090102.txt",
-//   "bankdata_20090203.txt",
-//   0 // Important: The array has to end with the 0 entry,
-//     // otherwise we have runaway and crash!
-// };
-#include "../bankdata/bankdatalist.h"
-// The array is now created at build time in the above file
+using namespace std;
 
 // A hard-coded replacement for strptime() because that one is not
 // available on windows/mingw.
-static void extract_date(const char *d, struct tm* p_tm)
+static void extract_date(const std::string& d, struct tm* p_tm)
 {
   char a[5];
 
@@ -101,6 +104,26 @@ static void extract_date(const char *d, struct tm* p_tm)
   //  cerr << "tm_mday=" << p_tm->tm_mday << " ";
 }
 
+/// Returns true if the given string starts with the given beginning
+static bool starts_with(const std::string& str, const std::string& beginning)
+{
+    return (str.size() >= beginning.size())
+        && (str.substr(0, beginning.size()) == beginning);
+}
+/// Returns false if the given string ends with the given ending.
+static bool ends_with(const std::string& str, const std::string& ending)
+{
+    return (str.size() >= ending.size())
+        && (str.substr(str.size() - ending.size(), str.size()) == ending);
+}
+/// Predicate for the file names that we accept
+static bool accept_filename(const std::string& f)
+{
+    return (f.size() == std::string("bankdata_20051234.txt").size())
+        && starts_with(f, "bankdata_")
+        && ends_with(f, ".txt");
+}
+
 #ifndef ONLY_EXTRACT_DATE_TEST
 const char* AccountNumberCheck::stringEncoding()
 {
@@ -162,31 +185,49 @@ std::string AccountNumberCheck::resultToString(Result r)
 AccountNumberCheck::AccountNumberCheck() 
     : data() // std::map doesn't take size as argument
 {
-   algorithms_init_binreloc();
-   populate_dated_files();
-   std::string datadir(bankdata_dir());
-   std::string dirsep(DIRSEP);
+   init_datafile_list();
 
-//    // Old implementation without dated bankdata
-//    std::string bankdata_filename = "bankdata.txt";
-//    std::string full_filename(
-//       datadir + dirsep + bankdata_filename);
-//    readDatedFile(make_pair( full_filename, make_pair(0,0)));
-
-   // New implementation; use dated bankdata files per default.
-   FileDaterange file = find_closest_datafile(time(NULL));
-   file.first = datadir + dirsep + file.first;
-   readDatedFile(file);
+   // New implementation; use dated bankdata files per
+   // default. However, this of course only works if we found the
+   // files beforehand.
+   if (!dated_files.empty())
+   {
+       assert(!dated_files.empty());
+       FileDaterange file = find_closest_datafile(time(NULL));
+       readDatedFile(file);
+   }
 }
 
 AccountNumberCheck::AccountNumberCheck(const string& filename) 
     : 
     data() // std::map doesn't take size as argument
 {
-  populate_dated_files();
+  init_datafile_list();
+
   readDatedFile(make_pair( filename, make_pair(0,0)));
 }
 
+void AccountNumberCheck::init_datafile_list()
+{
+   algorithms_init_binreloc();
+
+   populate_dated_files(bankdata_dir(), false);
+
+   // No files found? Maybe in the compiled-in installation location?
+   if (dated_files.empty())
+   {
+       populate_dated_files(BANKDATA_PATH, false);
+
+       // If there are still no files found, try again reading but now
+       // print the file-not-found error messages.
+       if (dated_files.empty())
+       {
+           populate_dated_files(bankdata_dir(), true);
+           populate_dated_files(BANKDATA_PATH, true);
+       }
+   }
+}
+
 AccountNumberCheck::~AccountNumberCheck() 
 {
   deleteList();
@@ -215,14 +256,18 @@ AccountNumberCheck::readFile(const string &filename)
   if (data.size() > 0)
     deleteList();
 
+  //cerr << "Will now read file: " << filename << std::endl;
+
   // Now read file
   FILE *istr = fopen(filename.c_str(), "r");
   // FIXME: Do a lot of error checking here.
   if (!istr)
   {
+     int errsv = errno;
+     std::string errstr = strerror(errsv);
      std::cerr << "AccountNumberCheck::readFile: File " << filename 
-	       << " could not be opened. "
-	"AccountNumberCheck could not obtain bank data." << std::endl;
+	       << " could not be opened: " << errstr
+               << "\nAccountNumberCheck could not obtain bank data." << std::endl;
      return;
   }
 
@@ -383,7 +428,79 @@ AccountNumberCheck::check(const string& bankId, const string& accountId,
 // ////////////////////////////////////////
 // Bankdata date validity
 
-void AccountNumberCheck::populate_dated_files()
+typedef std::vector<std::string> StringList;
+
+static StringList all_filenames_from_dir(const std::string& dir, bool print_errormsgs)
+{
+    StringList result;
+#ifdef HAVE_WINDOWS_H
+    HANDLE hFind = INVALID_HANDLE_VALUE;
+    WIN32_FIND_DATA ffd;
+    hFind = FindFirstFile(dir.c_str(), &ffd);
+    // List all the files in the directory with some info about them.
+
+    do
+    {
+        if (ffd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)
+        {
+            _tprintf(TEXT("  %s   <DIR>\n"), ffd.cFileName);
+        }
+        result.push_back(ffd.cFileName);
+    }
+    while (FindNextFile(hFind, &ffd) != 0);
+
+    FindClose(hFind);
+
+#else // HAVE_WINDOWS_H
+# ifdef HAVE_DIRENT_H
+    DIR *dp = opendir(dir.c_str());
+    if (dp)
+    {
+        struct dirent *ep;
+        while ( (ep = readdir(dp)) )
+        {
+            result.push_back(ep->d_name);
+            //cerr << "Found dir entry name " << result.back() << std::endl;
+        }
+        closedir(dp);
+    }
+    else
+    {
+        int errsv = errno;
+        std::string errstr = strerror(errsv);
+        if (print_errormsgs)
+            cerr << "Could not read directory \"" << dir << "\": " << errstr << std::endl;
+    }
+# else // HAVE_DIRENT_H
+#  error "OS is not yet supported: Neither POSIX <dirent.h> nor MS windows <windows.h> is available!"
+# endif // HAVE_DIRENT_H
+#endif // HAVE_WINDOWS_H
+    return result;
+}
+
+static StringList
+lookup_files_from_dir(const std::string& dir,
+                      bool (*filename_predicate) (const std::string&),
+                      bool print_errormsgs)
+{
+    // there is surely a STL algorithm for this, but I forgot which
+    // one it is (such as copy_if)
+    StringList raw = all_filenames_from_dir(dir, print_errormsgs);
+    StringList result;
+    for (StringList::const_iterator iter = raw.begin();
+         iter != raw.end(); ++iter)
+    {
+        if (filename_predicate(*iter))
+        {
+            result.push_back(*iter);
+            //cerr << "Adding entry name " << result.back() << std::endl;
+        }
+    }
+    return result;
+}
+
+
+void AccountNumberCheck::populate_dated_files(const std::string& dirname, bool print_errormessages)
 {
    dated_files.clear();
 
@@ -404,20 +521,36 @@ void AccountNumberCheck::populate_dated_files()
      unsetenv("TZ");
 #endif
 
-   // In the long run, this method should rather look up the available
-   // files from the bankdata directory. But for now we stick to the
-   // hard-coded list because at least we know the full date ranges.
-   const char **ptr = file_dates;
-   while (*ptr)
+   // This method now looks up the available files from the bankdata
+   // directory. Previously, we had a hard-coded list of available
+   // files.
+   StringList file_names = lookup_files_from_dir(dirname, &accept_filename,
+                                                 print_errormessages);
+   // Also sort alphabetically before using this
+   std::sort(file_names.begin(), file_names.end());
+
+   if (file_names.empty() && print_errormessages)
    {
-      string filename = *ptr;
-      ptr++;
+       std::cerr << "Oops, no bank data file was found in directory \"" << dirname
+                 << "\"! The ktoblzcheck library will not work." << std::endl;
+   }
+
+   StringList::const_iterator iter = file_names.begin();
+   while (iter != file_names.end())
+   {
+      string filename = *iter;
+      ++iter;
 
       //cerr << "Adding file " << filename << endl;
 
       // Use the part directly after the '_' character as the date
       struct tm tmp_tm;
-      extract_date(strchr(filename.c_str(), '_')+1, &tmp_tm);
+      {
+          const char* found_char = strchr(filename.c_str(), '_');
+          if (!found_char || *found_char == '\0')
+              break;
+          extract_date(std::string(found_char+1), &tmp_tm);
+      }
 
       // Point to the beginning of day (don't forget this as tmp_tm is
       // uninitialized so far!)
@@ -435,7 +568,7 @@ void AccountNumberCheck::populate_dated_files()
 	 // to crash because we couldn't find any bankdata. To avoid
 	 // this crash, we just make up a date interval in order to
 	 // use at least this very last file.
-	 if (dated_files.empty() && ! (*ptr))
+	 if (dated_files.empty() && iter == file_names.end())
 	 {
 	    time_t now = time(NULL);
 	    // We make up a date interval on our own: Let's guess the
@@ -447,7 +580,7 @@ void AccountNumberCheck::populate_dated_files()
 		      << ". Please report this error of ktoblzcheck "
 		      << KTOBLZCHECK_VERSION_MAJOR << "." << KTOBLZCHECK_VERSION_MINOR
 		      << " to aqbanking-devel@lists.sourceforge.net" << std::endl;
-	    dated_files.push_back(make_pair(string(filename),
+	    dated_files.push_back(make_pair(dirname + DIRSEP + string(filename),
 					    make_pair(start_date, end_date)));
 	 }
 	 continue;
@@ -461,7 +594,7 @@ void AccountNumberCheck::populate_dated_files()
 	 // end date of the previous entry.
 	 dated_files.back().second.second = start_date - 1;
       }
-      dated_files.push_back(make_pair(string(filename),
+      dated_files.push_back(make_pair(dirname + DIRSEP + string(filename),
 				      make_pair(start_date, end_date)));
    }
 
@@ -499,7 +632,9 @@ bool AccountNumberCheck::isValidDataAvailable(time_t date) const
    if (isDataValidForDate(date))
       return true;
 
-   assert(!dated_files.empty());
+   if (dated_files.empty())
+       return false;
+
    DatedFiles::const_iterator iter = dated_files.begin();
 
    // Is the given date already earlier than first starting date?
@@ -518,7 +653,9 @@ bool AccountNumberCheck::isValidDataAvailable(time_t date) const
 
 time_t AccountNumberCheck::closestValidData(time_t date) const
 {
-   assert(!dated_files.empty());
+   if (dated_files.empty())
+       return 0;
+
    DatedFiles::const_iterator iter = dated_files.begin();
 
    // Is the given date already earlier than first starting date?
@@ -538,10 +675,13 @@ time_t AccountNumberCheck::closestValidData(time_t date) const
 
 bool AccountNumberCheck::loadDataForDate(time_t date)
 {
+   if (dated_files.empty())
+       return false;
+
+   assert(!dated_files.empty());
    FileDaterange file = find_closest_datafile(date);
 
-   // Prepend the bankdata directory
-   file.first = bankdata_dir() + DIRSEP + file.first;
+   // Filename string has already prepended the bankdata directory
 
    readDatedFile(file);
    return isDataValidForDate(date);
@@ -559,12 +699,21 @@ void AccountNumberCheck::readDatedFile(const FileDaterange& filename)
 
 #else // ONLY_EXTRACT_DATE_TEST
 
+int rv = 0;
+
 // This is a small unittest for the extract_date() function at the
 // top, because this seems to fail on Mac OS X and this should at
 // least be discovered by make check.
 int main(int argc, char **argv) {
 
-    std::string filename = *file_dates;
+    std::string filename = "bankdata_20090102.txt";
+
+    if (!starts_with(filename, "bankdata_"))
+        rv = -1;
+    if (!ends_with(filename, ".txt"))
+        rv = -1;
+    if (!accept_filename(filename))
+        rv = -1;
 
     // Use the part directly after the '_' character as the date
     struct tm tmp_tm;
@@ -589,9 +738,9 @@ int main(int argc, char **argv) {
 	  "extract_date we got \"" << asctime(&tmp_tm) << "\" and from "
 #endif
 	  "mktime() we got " << ctime (&start_date) << endl;
-	return -1;
+        rv = -1;
     }
-    return 0;
+    return rv;
 }
 #endif // ONLY_EXTRACT_DATE_TEST
 
