@@ -1306,10 +1306,14 @@ AccountNumberCheck::Result method_B5(const int *account, int *weight) {
 }
 // B6, B7 and B8 Added by Alexander Kurz
 // new method as of September 5th, 2005, checked with Bundesbank-Testnumbers
+// method B6 was last changed for due-date September, 5 2011
 AccountNumberCheck::Result method_B6(const int *account, int *weight,
 				     const std::string& accountId, const std::string& bankId) {
     if (account[0] > 0 && account[0] <= 9)
 	return method_20(account,weight);	// variant 1
+    if (account[0] == 0 && account[1] == 2 && account[2] == 6
+	&& account[3] == 9 && ( account[4] > 0 || account[4] <= 9))
+	return method_20(account,weight);	// variant 1
     if (account[0] == 0)
 	return method_53(account, weight, accountId, bankId);	// variant 2
     return AccountNumberCheck::ERROR; // should never happen, to be on the safe side
@@ -1491,9 +1495,9 @@ AccountNumberCheck::Result method_D0(const int *account, int *weight) {
 	return AccountNumberCheck::OK;				// variant 2, "method 9" again :-))
 }
 // method D1 was modified with due date starting from 7. June 2010
-// method D1 was last modified for due date 7. March 2011
+// method D1 was last modified for due date 5. September 2011
 AccountNumberCheck::Result method_D1(const int *account, int *weight) {
-    if (account[0] == 2 || account[0] == 7 || account[0] == 8)
+    if (account[0] == 7 || account[0] == 8)
 	return AccountNumberCheck::ERROR;
     // note: this method is quite similar to method_C6
     number2Array("2121212120", weight);
